unit AHListBox;
{
  TAHListBox version 1.0 10/02/99

  Author: Francesc Puigdemont Casamaj
  Copyright  1999 AdHoc Software.

  For questions and suggestions eMail to amer@lix.intercom.es

  Delphi version: 2, 3

  Description
    TCustomListBox descendant with images and text align.
    You can set the selection bar with three styles: Raised, Lowered or Color.
    The images are handled with an ImageList and OnGetImageIndex event and can
    be positioned to left or right.
    Only horizontal text alignment, vertical alignment is always centered.
    No multi lines in this version.
    Odd / Even colors.
    You can set an offset for the text.

  This component is POSTCARDWARE, so send me postcards from your city, country
  or whatever you want to:

  Francesc Puigdemont
  C/ Sant Miquel, 6
  17170 - AMER - Girona
  Catalonia

  "AS-IS"
  No warranty is given by the author, expressed or implied.
  Do not distribute modified version without my agreement.
  Thanks and sorry for my (bad) english (aah! this is english!).
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, StdCtrls, ExtCtrls;

type
  TBarStyle = (tbsLowered, tbsRaised, tbsColor);
  TImagePos = (ipLeft, ipRight);

  TAHGetIndexEvent = procedure(Sender: TObject; const ItemIndex: Integer; var idx: integer) of object;

  TAHListBox = class(TCustomListBox)
  private
    { Private declarations }
    FBarFont:         TFont;
    FBarColor:        TColor;
    FOffset:          Byte;
    FBarStyle:        TBarStyle;
    FImagePos:        TImagePos;
    FAlignment:       TAlignment;
    FImages:          TImageList;
    FImageChangeLink: TChangeLink;
    FOnGetImageIndex: TAHGetIndexEvent;
    FOddEvenColors:   Boolean;
    FOddColor:        TColor;
    FEvenColor:       TColor;
    FOnDrawItem: TDrawItemEvent;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure SetBarFont( F: TFont );
    procedure SetBarColor( C: TColor );
    procedure SetOffset( b: Byte );
    procedure SetBarStyle( S: TBarStyle );
    procedure SetAlignment(Value : TAlignment);
    procedure SetImagePos( ip: TImagePos );
    procedure BarFontChange( Sender: TObject );
    procedure ImageListChange (Sender: TObject);
    procedure SetImages (Value: TImageList);
    procedure CalcHeight;
    //-
    procedure SetOddEvenColors( b: Boolean );
    procedure SetOddColor( C: TColor );
    procedure SetEvenColor( C: TColor );
  protected
    { Protected declarations }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState); override;

  published
    { Published declarations }
    property Align;
    //-
    property Alignment : TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property BarColor: TColor    read FBarColor write SetBarColor;
    property BarFont:  TFont     read FBarFont  write SetBarFont;
    property BarStyle: TBarStyle read FBarStyle write SetBarStyle;
    //-
    property BorderStyle;
    property Color;
    property Columns;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    //-
    property EvenColor: TColor read FEvenColor write SetEvenColor;
    //-
    property ExtendedSelect;
    property Font;
    //-
    property Images: TImageList read FImages write SetImages;
    property ImagePos: TImagePos read FImagePos write SetImagePos default ipLeft;
    //-
    property IntegralHeight;
    property ItemHeight;
    property Items;
    property MultiSelect;
    //-
    property OddEvenColors:  Boolean read FOddEvenColors  write SetOddEvenColors  default false;
    property OddColor:       TColor  read FOddColor       write SetOddColor;
    property Offset:         Byte    read FOffset         write SetOffset         default 3;
    //-
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabWidth;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    //-
    property OnGetImageIndex: TAHGetIndexEvent read FOnGetImageIndex write FOnGetImageIndex;
    //-
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

const V_OFFSET = 5;

procedure Register;

implementation

{$R *.DCR}

function Max( a, b: Integer ): Integer;
begin
  Result := a;
  if b > a then result := b;
end;

Constructor TAHListBox.Create(AOwner: TComponent);
begin
 inherited Create(AOwner);
 Style          := lbOwnerDrawFixed;
 ItemHeight     := 20;
 Color          := $00D6E9D7;
 FBarColor      := $00D6E9D7;
 FBarStyle      := tbsLowered;
 FBarFont       := TFont.Create;
 FBarFont.Name  := Font.Name;
 FBarFont.Size  := 8;
 FBarFont.Color := clRed;
 FOffset        := 3;
 FAlignment     := taLeftJustify;
 FImagePos      := ipLeft;

 //-
 FOddEvenColors := False;
 FOddColor      := $00d1d1d1;
 FEvenColor     := $00e1e1e1;
 //-

 FBarFont.OnChange         := BarFontChange;
 FImageChangeLink          := TChangeLink.Create;
 FImageChangeLink.OnChange := ImageListChange;
end;

destructor TAHListBox.Destroy;
begin
 FBarFont.Free;
 inherited Destroy;
end;

procedure TAHListBox.CalcHeight;
var h: Integer;
begin
  Canvas.Font := Font;
  if Assigned(FImages) then
     h := Max( Canvas.TextHeight( 'y' ), FImages.Height )
 else
     h := Canvas.TextHeight('y');

 Canvas.Font := FBarFont;
 h := Max( h, Canvas.TextHeight( 'y' ) ) + V_OFFSET;
 if ItemHeight <> h then ItemHeight := h;
 Canvas.Font := Font;
end;

procedure TAHListBox.CMFontChanged(var Message: TMessage);
begin
 inherited;
 Canvas.Font := Font;
 CalcHeight;
end;

procedure TAHListBox.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = Images then Images := nil;
  end;
end;

procedure TAHListBox.ImageListChange(Sender: TObject);
begin
  CalcHeight;
end;

procedure TAHListBox.SetImages (Value: TImageList);
begin
  if Assigned (FImages) then
    Images.UnRegisterChanges(FImageChangeLink);

  FImages := Value;

  if Assigned (FImages) then
     Images.RegisterChanges (FImageChangeLink);

  CalcHeight;
  Invalidate;
end;

procedure TAHListBox.SetImagePos( ip: TImagePos );
begin
  if ip <> FImagePos then begin
    FImagePos := ip;
    if Assigned(FImages) then Invalidate;
  end;
end;

procedure TAHListBox.SetAlignment(Value : TAlignment);
begin

  if value <> FAlignment then
  begin
    FAlignment := Value;
    Invalidate;
  end;

end;

procedure TAHListBox.SetBarFont( F: TFont );
begin
 if FBarFont <> F then begin
    FBarFont.Assign( F );
    CalcHeight;
 end;
end;

procedure TAHListBox.BarFontChange(Sender: TObject);
var r: TRect;
begin
  CalcHeight;
  if ItemIndex > -1 then begin
     if MultiSelect and (SelCount > 1) then invalidate
     else begin
       r := ItemRect( ItemIndex );
       InvalidateRect( Handle, @r, false );
     end;
   end;

//  Invalidate;
end;

procedure TAHListBox.SetBarColor( C: TColor );
var r: TRect;
begin
  if c <> FBarColor then begin
     FBarColor := c;
     if ItemIndex > -1 then begin
        if MultiSelect and (SelCount > 1) then invalidate
        else begin
          r := ItemRect( ItemIndex );
          InvalidateRect( Handle, @r, false );
        end;
     end;
  end;
end;

procedure TAHListBox.SetOffset( b: Byte );
begin
    if b <> FOffset then begin
      FOffset := b;
      invalidate;
    end;
end;

procedure TAhListBox.SetBarStyle( S: TBarStyle );
var r: TRect;
begin
   if s <> FBarStyle then begin
     FBarStyle := S;
     if ItemIndex > -1 then begin
        if MultiSelect and (SelCount > 1) then invalidate
        else begin
          r := ItemRect( ItemIndex );
          InvalidateRect( Handle, @r, false );
        end;
     end;
   end;
end;
//-
procedure TAhListBox.SetOddEvenColors( B: Boolean );
begin
   if b <> FOddEvenColors then begin
      FOddEvenColors := b;
      Invalidate;
   end;
end;

procedure TAhListBox.SetOddColor( C: TColor );
begin
   if c <> FOddColor then begin
      FOddColor := C;
      Invalidate;
   end;
end;

procedure TAhListBox.SetEvenColor( C: TColor );
begin
   if c <> FEvenColor then begin
      FEvenColor := C;
      Invalidate;
   end;
end;

//-
procedure TAHListBox.DrawItem( Index: Integer; Rect: TRect; State: TOwnerDrawState);
Const
  TextAlignments : Array[TAlignment] of Word = (dt_Left, dt_Right, dt_Center);

var XR     : Trect;
   opcions : integer;
   ImgIdx  : Integer;
begin

 if Assigned(FOnDrawItem) then FOnDrawItem( Self, Index, Rect, State ) else
 begin
  ImgIdx := -1;

  if Assigned(FImages) then
    if Assigned(FOnGetImageIndex) then FOngetImageIndex(Self, Index, ImgIdx );

  With Canvas do Begin
    if (odSelected in State) then  Brush.Color :=  FBarColor
    else if FOddEvenColors then
      if odd(Index) then Brush.Color := FOddColor
      else Brush.Color := FEvenColor
    else
      Brush.Color := Color;

    fillRect(rect);

    Pen.Mode := pmCopy;
    Pen.Width := 1;
    Pen.Style := psSolid;

    xr := Rect;

    If (odSelected in State) then Begin
      if FBarStyle = tbsRaised then
         Frame3d(Canvas, Rect, clBtnShadow, clBtnHighLight, 1 )
      else if FBarStyle = tbsLowered then
         Frame3d(Canvas, Rect, clBtnHighLight, clBtnShadow, 1 )
      else begin
         Pen.Color := FBarColor;
         Rectangle( Rect.Left, Rect.Top, Rect.Right, Rect.Bottom );
      end;
    end else begin
      inc(Rect.Left); inc(Rect.Top);
      Dec(Rect.Bottom); dec(Rect.Right);
    end;

    if FImages <> nil then begin
      if (ImgIdx > -1) then
        if FImagePos = ipLeft then begin
         FImages.Draw(Canvas, Rect.Left + FOffset, Rect.Top +
                      ((Rect.Bottom - Rect.Top) div 2) - FImages.Height div 2,
                      ImgIdx );
         Rect.Left := Rect.Left + FOffset + FImages.Width + 3;
        end else begin
         FImages.Draw(Canvas, Rect.Right - FOffset - FImages.Width, Rect.Top +
                      ((Rect.Bottom - Rect.Top) div 2) - FImages.Height div 2,
                      ImgIdx );
         Rect.Right := Rect.Right - FOffset - FImages.Width - 3;
        end;
        if (FAlignment = taLeftJustify) and (FImagePos = ipRight) then
            inc(Rect.Left, FOffset)
        else inc(Rect.Left, 3);
        if (FAlignment = taRightJustify) and (FImagePos = ipLeft) then
            dec(Rect.Right, FOffset)
        else dec(Rect.Right, 3);
    end else begin
      if FAlignment = taLeftJustify then  Rect.Left := Rect.Left + FOffset;

      if FAlignment = taRightJustify then Rect.Right := Rect.Right - FOffset - 1
      else Dec(Rect.Right);

    end;

    Dec(Rect.Bottom);

    If (odSelected in State) or(odFocused in State) then Font := FBarFont;

    opcions := TextAlignments[FAlignment] or DT_SINGLELINE or DT_VCENTER;

    setbkMode(Canvas.Handle, Transparent);
    DrawText(Canvas.Handle, PChar(Items[Index]), -1, Rect, opcions);

    if (odFocused in State) and (odselected in state) then
      DrawFocusRect(xr);

  End;
 end;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TAHListBox]);
end;

end.
